///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Contains all memory-related code.
 *	\file		IceMemoryManager.h
 *	\author		Pierre Terdiman
 *	\date		April, 4, 2000
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Include Guard
#ifndef __ICEMEMORYMANAGER_H__
#define __ICEMEMORYMANAGER_H__

#undef ZeroMemory
#undef CopyMemory
#undef MoveMemory
#undef FillMemory

//!	A function to clear a buffer.
//!	\param	addr		buffer address
//!	\param	size		buffer length
//!	\see	FillMemory
//!	\see	CopyMemory
//!	\see	MoveMemory
ICECORE_API __forceinline void ZeroMemory(void* addr, udword size)					{ memset(addr, 0, size);		}

//!	A function to fill a buffer with a given byte.
//!	\param	addr		buffer address
//!	\param	size		buffer length
//!	\param	val			the byte value
//!	\see	ZeroMemory
//!	\see	CopyMemory
//!	\see	MoveMemory
ICECORE_API __forceinline void FillMemory(void* dest, udword size, ubyte val)		{ memset(dest, val, size);		}

//!	A function to copy a buffer.
//!	\param	addr		destination buffer address
//!	\param	addr		source buffer address
//!	\param	size		buffer length
//!	\see	ZeroMemory
//!	\see	FillMemory
//!	\see	MoveMemory
ICECORE_API __forceinline void CopyMemory(void* dest, const void* src, udword size)	{ memcpy(dest, src, size);		}

//!	A function to move a buffer.
//!	\param	addr		destination buffer address
//!	\param	addr		source buffer address
//!	\param	size		buffer length
//!	\see	ZeroMemory
//!	\see	FillMemory
//!	\see	CopyMemory
ICECORE_API __forceinline void MoveMemory(void* dest, const void* src, udword size)	{ memmove(dest, src, size);		}

ICECORE_API udword GetFreeRam();

#ifdef _DEBUG
ICECORE_API void* Alloc(udword size, const char* file, sdword line);
ICECORE_API bool Free(void* addr);
#else
ICECORE_API void* Alloc(udword size);
ICECORE_API void Free(void* addr);
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MACROS
#define SIZEOFOBJECT		sizeof(*this)								//!< Gives the size of current object. Avoid some mistakes (e.g. "sizeof(this)").
//#define CLEAROBJECT		{ memset(this, 0, SIZEOFOBJECT);	}		//!< Clears current object. Laziness is my business. HANDLE WITH CARE.
#define DELETESINGLE(x)		{ if (x) delete x; x = null;		}		//!< Deletes an instance of a class.
#define DELETEARRAY(x)		{ if (x) delete []x; x = null;		}		//!< Deletes an array.
#define CHECKALLOC(x)		{ if(!x) { SetIceErrorMsg("Out of memory.", 1); return false;} }	//!< Standard alloc checking. HANDLE WITH CARE.

#ifdef _DEBUG
#define MEMBLOCKSTART		64
inline void* __cdecl operator new(unsigned int size, const char* file, sdword line)	{	return Alloc(size, file, line);	}	//!< new overload
inline void operator delete(void* addr, const char* file, sdword line)				{	Free(addr);						}	//!< delete overload
inline void operator delete(void* addr)												{	Free(addr);						}	//!< delete overload
#undef NEW_D
#undef DELETE_D
#define NEW_D ::new(THIS_FILE, __LINE__)
#define DELETE_D ::delete
#define new NEW_D
#define delete DELETE_D
#else
#ifndef DEFAULT_NEWDELETE
inline void* __cdecl operator new(unsigned int size)								{	return Alloc(size);				}	//!< new overload
inline void operator delete(void* addr)												{	Free(addr);						}	//!< delete overload
#endif
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//! The memory manager contains a linked list of memory cells.
	typedef struct ICECORE_API MemoryCell {
		struct MemoryCell*	PreviousCell;				//!< Link to previous cell or null
		struct MemoryCell*	NextCell;					//!< Link to next cell or null
		void*				MallocAddress;				//!< Address of allocated ram
		void*				AlignedMallocAddress;		//!< Aligned address (on 1K bytes boundary)
		udword				ReservedBytes;				//!< Size of allocated ram
	} MEMORY_CELL;

	class ICECORE_API MemoryManager
	{
		private:

		MEMORY_CELL*		mCurrentCell;				//!< Active cell
		MEMORY_CELL*		mInitCell;					//!< Initial cell
		long				mTotal;						//!< Total allocated ram
		bool				mIsAligned;					//!< Is alignment on/off ?

		public:

		//! Constructor
		MemoryManager();
		//! Destructor
		~MemoryManager();

		void*				Malloc(size_t n);								//!< Allocates a buffer and keep track of it.
		bool				Free(void* buffer);								//!< Frees a previously allocated buffer.
		long				RamUsed()			{ return mTotal;		}	//!< Returns total ram used
		void				AlignOn()			{ mIsAligned = true;	}	//!< Set the alignment on
		void				AlignOff()			{ mIsAligned = false;	}	//!< Set the alignment off
	};

#endif // __ICEMEMORYMANAGER_H__
