///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	This file contains the ASCII Format exporter plug-in for Flexporter.
 *	\file		IceASCIIExporter.cpp
 *	\author		Pierre Terdiman
 *	\date		April, 4, 2000
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Precompiled Header
#include "Stdafx.h"

static ASCIIFormat		gASCIIExporter;		// Gloabl exporter instance
static ExportSettings	gSettings;			// Global export settings

// FLEXPORTER Identification Callbacks

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Gives a brief exporter description. That string is displayed inside MAX, in the Options Panel.
 *	\relates	ASCIIFormat
 *	\fn			ExporterDescription()
 *	\return		a description string
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT const char* ExporterDescription()
{ 
	return "ICE ASCII Exporter";
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Returns the format's extension. That string is displayed inside MAX, in the Options Panel.
 *	\relates	ASCIIFormat
 *	\fn			FormatExtension()
 *	\return		an extension string
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT const char* FormatExtension()
{
	return "txt";
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Returns the one and only exporter instance.
 *	\relates	ASCIIFormat
 *	\fn			GetExporter()
 *	\return		pointer to the global exporter instance.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT ExportFormat* GetExporter()
{
	return &gASCIIExporter;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Returns the default exportation settings for this format. This is the right place to initialize the default settings for your format.
 *	\relates	ASCIIFormat
 *	\fn			GetDefaultSettings()
 *	\return		pointer to the global ExportSettings instance.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT ExportSettings* GetDefaultSettings()
{
	// General settings
	gSettings.Compressed					= false;
	gSettings.ExportWholeScene				= true;
	gSettings.ExportVisTrack				= false;
	gSettings.ExportHiddenNodes				= false;
	// Mesh settings
	gSettings.ExpUVW						= true;
	gSettings.DiscardW						= true;
	gSettings.ExpVTXColor					= true;
	gSettings.UseSmgrp						= true;
	gSettings.RemoveScaling					= true;
	gSettings.ConvertToD3D					= true;
	gSettings.AbsolutePRS					= false;
	gSettings.ConvexHull					= false;
	gSettings.BoundingSphere				= false;
	gSettings.InertiaTensor					= false;
	// Consolidation settings
	gSettings.ConsolidateMesh				= false;
	gSettings.ComputeFaceNormals			= false;
	gSettings.ComputeVertexNormals			= true;
	gSettings.ExportNormalInfo				= false;
	// Material settings
	gSettings.ForceAmbient					= false;
	gSettings.ForceDiffuse					= false;
	// Texture settings
	gSettings.OpacityInAlpha				= false;
	gSettings.TexnameOnly					= true;
	gSettings.KeepAmbientTexture			= false;
	gSettings.KeepDiffuseTexture			= true;
	gSettings.KeepSpecularTexture			= false;
	gSettings.KeepShininessTexture			= false;
	gSettings.KeepShiningStrengthTexture	= false;
	gSettings.KeepSelfIllumTexture			= false;
	gSettings.KeepOpacityTexture			= true;
	gSettings.KeepFilterTexture				= false;
	gSettings.KeepBumpTexture				= false;
	gSettings.KeepReflexionTexture			= false;
	gSettings.KeepRefractionTexture			= false;
	gSettings.KeepDisplacementTexture		= false;
	gSettings.TexMaxSize					= 0;
	// Camera settings
	gSettings.ExpCameras					= true;
	gSettings.ExpTargetNodes				= true;
	// Light settings
	gSettings.ExpLights						= true;
	// Animation settings
	gSettings.SingleFrame					= true;
	gSettings.Sampling						= true;
	gSettings.SamplingRate					= 1;

	return &gSettings;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Returns the FLEXPORTER SDK Version.
 *	\relates	ASCIIFormat
 *	\fn			Version()
 *	\return		FLEXPORTER_VERSION
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT int Version()
{
	return FLEXPORTER_VERSION;
}










///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	An ASCII exporter plug-in for Flexporter..
 *	\class		ASCIIFormat
 *	\author		Pierre Terdiman
 *	\version	1.0
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Constructor.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
ASCIIFormat::ASCIIFormat()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Destructor.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
ASCIIFormat::~ASCIIFormat()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Plug-in initialization method.
 *	This method is called once before each export. When this method is called, the mSettings and mFilename members of the base format are valid.
 *	\param		true for pure motion files.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::Init(bool motion)
{
	mGeneral
		.StoreASCII("// File created by FLEXPORTER\n")
		.StoreASCII("// Copyright (c) 2000 Pierre Terdiman\n");

	if(!motion)
	{
		mGeneral		.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("GENERAL INFORMATION\n");
		mGeomObjects	.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("MESHES\n");
		mCameras		.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("CAMERAS\n");
		mLights			.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("LIGHTS\n");
		mShapes			.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("SHAPES\n");
		mHelpers		.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("HELPERS\n");
		mTexmaps		.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("TEXTURES\n");
		mMaterials		.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("MATERIALS\n");
		mControllers	.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("CONTROLLERS\n");
	}
	mMotion			.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("MOTIONS\n");

	mIsMotionFile	= motion;

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Scene export method.
 *	This method is called once to export the scene global information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::SetSceneInfo(MainInfo& maininfo)
{
	// Save time info for later
	CopyMemory(&mTimeInfo, &maininfo.Time, sizeof(TimeInfo));

	// Export time info
	mGeneral
		.StoreASCII("First frame: ")		.StoreASCII((long)(mTimeInfo.StartTime / mTimeInfo.DeltaTime))
		.StoreASCII("\nLast frame: ")		.StoreASCII((long)(mTimeInfo.EndTime / mTimeInfo.DeltaTime))
		.StoreASCII("\nFrame rate: ")		.StoreASCII(mTimeInfo.FrameRate)
		.StoreASCII("\nTicks per frame: ")	.StoreASCII((long)mTimeInfo.DeltaTime)
		.StoreASCII("\n");

	// Export background color
	mGeneral
		.StoreASCII("Background color: ")
		.StoreASCII(maininfo.BackColor.r).StoreASCII(" ")
		.StoreASCII(maininfo.BackColor.g).StoreASCII(" ")
		.StoreASCII(maininfo.BackColor.b).StoreASCII("\n");

	// Export ambient color
	mGeneral
		.StoreASCII("Global ambient color: ")
		.StoreASCII(maininfo.AmbColor.r).StoreASCII(" ")
		.StoreASCII(maininfo.AmbColor.g).StoreASCII(" ")
		.StoreASCII(maininfo.AmbColor.b).StoreASCII("\n");

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Common method used to export basic information for a mesh, a camera, a light, etc.
 *	\param		obj			current object
 *	\param		array		destination export array
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportBasicInfo(ObjectInfo* obj, CustomArray* array)
{
	if(!array || !obj) return false;

	// Export name
	array	->StoreASCII((const char*)obj->mName);

	// Export database management information
	array	->StoreASCII("\nNodeID:       ").StoreASCII(obj->mObjectID)
			.StoreASCII("\nParentNodeID: ")	.StoreASCII(obj->mParentID)
			.StoreASCII("\nLinkID:       ")	.StoreASCII(obj->mLinkID)
			.StoreASCII("\nIsInGroup:    ")	.StoreASCII(obj->mIsGroupMember)
			.StoreASCII("\n");

	// Export common properties
	array	->StoreASCII("\nPosition:     ").StoreASCII(obj->mPrs.Position.x)	.StoreASCII("  ")
											.StoreASCII(obj->mPrs.Position.y)	.StoreASCII("  ")
											.StoreASCII(obj->mPrs.Position.z)

			.StoreASCII("\nRotation:     ")	.StoreASCII(obj->mPrs.Rotation.x)	.StoreASCII("  ")
											.StoreASCII(obj->mPrs.Rotation.y)	.StoreASCII("  ")
											.StoreASCII(obj->mPrs.Rotation.z)	.StoreASCII("  ")
											.StoreASCII(obj->mPrs.Rotation.w)

			.StoreASCII("\nScale:        ")	.StoreASCII(obj->mPrs.Scale.x)		.StoreASCII("  ")
											.StoreASCII(obj->mPrs.Scale.y)		.StoreASCII("  ")
											.StoreASCII(obj->mPrs.Scale.z)

			.StoreASCII("\nWireColor:    ")	.StoreASCII(obj->mWireColor)

			.StoreASCII("\nLocal PRS:    ")	.StoreASCII(obj->mLocalPRS)

			.StoreASCII("\nD3D compliant:")	.StoreASCII(obj->mD3DCompliant);

	// Export user-defined properties
	if(obj->mUserProps)	array->StoreASCII("\nUser props:   ").StoreASCII((const char*)obj->mUserProps);

	array->StoreASCII("\n\n");

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Camera export method.
 *	This method is called once for each exported camera.
 *	\param		camera		a structure filled with current camera information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportCamera(CameraInfo& camera)
{
	// Export common information
	mCameras.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("Camera:       ");
	ExportBasicInfo(&camera, &mCameras);

	// Export camera parameters
	if(camera.mOrthoCam)
	{
		mCameras.StoreASCII("Type: orthographic\nWidth: ").StoreASCII(camera.mFOV).StoreASCII("\n");
	}
	else
	{
		mCameras.StoreASCII("Type: perspective\nField Of View: ").StoreASCII(camera.mFOV).StoreASCII("\n");
	}

	// Horizon line display
	mCameras.StoreASCII("Horizon line display: ")	.StoreASCII(camera.mHLineDisplay).StoreASCII("\n");				// Horizon line display

	// Export environment ranges if needed
	// Environment ranges
	mCameras
		.StoreASCII("\nENVIRONMENT RANGES:\n")
		.StoreASCII("\nNear range: ")				.StoreASCII(camera.mEnvNearRange)								// Near range
		.StoreASCII("\nFar range: ")				.StoreASCII(camera.mEnvFarRange)								// Far range
	// Environment display
		.StoreASCII("\nEnv display: ")				.StoreASCII(camera.mEnvDisplay)									// Environment display
		.StoreASCII("\n");

	// Export clipping planes if needed
	mCameras
		.StoreASCII("\nCLIPPING PLANES:\n")
		.StoreASCII("\nManual clip: ")				.StoreASCII(camera.mManualClip)									// Manual clip
		.StoreASCII("\nNear clip: ")				.StoreASCII(camera.mNearClip)									// Near clip
		.StoreASCII("\nFar clip: ")					.StoreASCII(camera.mFarClip)									// Far clip
		.StoreASCII("\n");

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Controller export method.
 *	This method is called once for each exported controller.
 *	\param		controller		a structure filled with current controller information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportController(ControllerInfo& controller)
{
	// Export database management information
	mControllers.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("Controller: ");
	mControllers
		.StoreASCII((const char*)controller.mField)
		.StoreASCII("\nController ID: ").StoreASCII(controller.mObjectID)
		.StoreASCII("\nOwner ID: ").StoreASCII(controller.mOwnerID);

	// Controller data
	ControllerData* cdata = controller.mData;

	mControllers
		.StoreASCII("\nController data: \n");

	udword NbValues;

			if(cdata->mType==CTRL_FLOAT)		{mControllers.StoreASCII("Float controller, ");			NbValues = 1; }
	else	if(cdata->mType==CTRL_VECTOR)		{mControllers.StoreASCII("Vector controller, ");		NbValues = 3; }
	else	if(cdata->mType==CTRL_QUAT)			{mControllers.StoreASCII("Quaternion controller, ");	NbValues = 4; }
	else	if(cdata->mType==CTRL_PR)			{mControllers.StoreASCII("PR controller, ");			NbValues = 7; }
	else	if(cdata->mType==CTRL_PRS)			{mControllers.StoreASCII("PRS controller, ");			NbValues = 10; }
	else										{mControllers.StoreASCII("Unknown controller, ");		NbValues = 0; }

			if(cdata->mMode==CTRL_SAMPLES)		mControllers.StoreASCII("sampling mode.\n\n");
	else	if(cdata->mMode==CTRL_KEYFRAMES)	mControllers.StoreASCII("keyframing mode.\n\n");
	else										mControllers.StoreASCII("Unknown mode.\n\n");

	// Dump the samples
	if(cdata->mMode==CTRL_SAMPLES)
	{
		SampleData* sdata = (SampleData*)cdata;
		mControllers.StoreASCII(sdata->mNbSamples).StoreASCII(" samples, sampling rate = ").StoreASCII(sdata->mSamplingRate).StoreASCII("\n");

		float* Values = (float*)sdata->mSamples;
		for(udword i=0;i<NbValues*sdata->mNbSamples;i++)
		{
			mControllers.StoreASCII(*Values++).StoreASCII(" ");
			if((i&63)==63)	mControllers.StoreASCII("\n");
		}
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Helper export method.
 *	This method is called once for each exported helper.
 *	\param		helper		a structure filled with current helper information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportHelper(HelperInfo& helper)
{
	// Export common information
	mHelpers.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("Helper:       ");
	ExportBasicInfo(&helper, &mHelpers);

	// Export helper parameters
	mHelpers.StoreASCII("Group head:  ").StoreASCII(helper.mIsGroupHead).StoreASCII("\n");

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Light export method.
 *	This method is called once for each exported light.
 *	\param		light		a structure filled with current light information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportLight(LightInfo& light)
{
	// Export common information
	mLights.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("Light:       ");
	ExportBasicInfo(&light, &mLights);

	// Export light parameters

	// Get light type
	mLights.StoreASCII("Type: ");
	if (light.mLightType==LTYPE_OMNI)					mLights.StoreASCII("omnidirectional light");
	else if (light.mLightType==LTYPE_TSPOT)				mLights.StoreASCII("targeted spot light");
	else if (light.mLightType==LTYPE_DIR)				mLights.StoreASCII("directional light");
	else if (light.mLightType==LTYPE_FSPOT)				mLights.StoreASCII("free spot light");
	else if (light.mLightType==LTYPE_TDIR)				mLights.StoreASCII("targeted directional light");
	mLights.StoreASCII("\n");

	// Color
	{
		mLights
			.StoreASCII("Intensity (Multiplier): ")			.StoreASCII(light.mIntensity)					// Intensity (multiplier)
			.StoreASCII("\nContrast: ")						.StoreASCII(light.mContrast)					// Contrast
			.StoreASCII("\nDiffuse Soft: ")					.StoreASCII(light.mDiffuseSoft)					// Diffuse soft
			.StoreASCII("\nColor: ")						.StoreASCII(light.mColor.x)						// Color
			.StoreASCII(" ")								.StoreASCII(light.mColor.y)						// Color
			.StoreASCII(" ")								.StoreASCII(light.mColor.z)						// Color
			.StoreASCII("\nUsed: ")							.StoreASCII((BOOL)light.mLightUsed)				// Light used or not
			.StoreASCII("\nAffect diffuse: ")				.StoreASCII((BOOL)light.mAffectDiffuse)			// Affect diffuse
			.StoreASCII("\nAffect specular: ")				.StoreASCII((BOOL)light.mAffectSpecular)		// Affect specular
			.StoreASCII("\n");
	}

	// Attenuation parameters
	{
		mLights
			.StoreASCII("Use near atten: ")					.StoreASCII((BOOL)light.mUseAttenNear)			// Use near atten
			.StoreASCII("\nShow near atten: ")				.StoreASCII((BOOL)light.mAttenNearDisplay)		// Show near atten
			.StoreASCII("\nNearAttenStart: ")				.StoreASCII(light.mNearAttenStart)				// Near atten start
			.StoreASCII("\nNearAttenEnd: ")					.StoreASCII(light.mNearAttenEnd)				// Near atten end
			.StoreASCII("\nUseAtten: ")						.StoreASCII(light.mUseAtten)					// Use atten
			.StoreASCII("\nShowAtten: ")					.StoreASCII(light.mShowAtten)					// Show atten
			.StoreASCII("\nAttenStart: ")					.StoreASCII(light.mAttenStart)					// Atten start
			.StoreASCII("\nAttenEnd: ")						.StoreASCII(light.mAttenEnd)					// Atten end
			.StoreASCII("\nDecay type: ")					.StoreASCII(light.mDecayType)					// Decay type
			.StoreASCII("\n");
	}

	// Spotlight parameters or directional parameters
	{
		if(light.mIsSpot)							mLights.StoreASCII("\nSPOTLIGHT PARAMETERS:\n\n");
		else if(light.mIsDir)						mLights.StoreASCII("\nDIRECTIONAL PARAMETERS:\n\n");

		if(light.mLightType!=LTYPE_OMNI)
		{
			mLights
				.StoreASCII("Hotspot: ")					.StoreASCII(light.mHotSpot)						// Hotspot
				.StoreASCII("\nFalloff: ")					.StoreASCII(light.mFallsize)					// Falloff
				.StoreASCII("\nAspect: ")					.StoreASCII(light.mAspect)						// Aspect
				.StoreASCII("\nSpot shape: ");																// Spot shape (RECT_LIGHT, CIRCLE_LIGHT)
					if(light.mSpotShape==SPOTSHP_RECT)		mLights.StoreASCII("rectangle\n");
			else	if(light.mSpotShape==SPOTSHP_CIRCLE)	mLights.StoreASCII("circle\n");
			else											mLights.StoreASCII("unknown!\n");
			mLights
				.StoreASCII("\nOvershoot: ")				.StoreASCII(light.mOvershoot)					// Overshoot
				.StoreASCII("\nShow cone: ")				.StoreASCII(light.mConeDisplay)					// Show cone
				.StoreASCII("\nTarget distance: ")			.StoreASCII(light.mTDist)						// Target distance
				.StoreASCII("\n");
		}
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Material export method.
 *	This method is called once for each exported material.
 *	\param		material		a structure filled with current material information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportMaterial(MaterialInfo& material)
{
	// Export database management information
	mMaterials.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("Material: ");
	mMaterials.StoreASCII((const char*)material.mName);


	// Export material parameters

	mMaterials
		.StoreASCII("\nMaterialID:              ").StoreASCII(material.mObjectID)
		.StoreASCII("\nAmbient map ID:          ").StoreASCII(material.mAmbientMapID)
		.StoreASCII("\nDiffuse map ID:          ").StoreASCII(material.mDiffuseMapID)
		.StoreASCII("\nSpecular map ID:         ").StoreASCII(material.mSpecularMapID)
		.StoreASCII("\nShininess map ID:        ").StoreASCII(material.mShininessMapID)
		.StoreASCII("\nShining Strength map ID: ").StoreASCII(material.mShiningStrengthMapID)
		.StoreASCII("\nSelfIllum map ID:        ").StoreASCII(material.mSelfIllumMapID)
		.StoreASCII("\nOpacity map ID:          ").StoreASCII(material.mOpacityMapID)
		.StoreASCII("\nFilter map ID:           ").StoreASCII(material.mFilterMapID)
		.StoreASCII("\nBump map ID:             ").StoreASCII(material.mBumpMapID)
		.StoreASCII("\nReflexion map ID:        ").StoreASCII(material.mReflexionMapID)
		.StoreASCII("\nRefraction map ID:       ").StoreASCII(material.mRefractionMapID)
		.StoreASCII("\nDisplacement map ID:     ").StoreASCII(material.mDisplacementMapID)

		.StoreASCII("\nAmbient Coeff:           ").StoreASCII(material.mAmbientCoeff)
		.StoreASCII("\nDiffuse Coeff:           ").StoreASCII(material.mDiffuseCoeff)
		.StoreASCII("\nSpecular Coeff:          ").StoreASCII(material.mSpecularCoeff)
		.StoreASCII("\nShininess Coeff:         ").StoreASCII(material.mShininessCoeff)
		.StoreASCII("\nShining Strength Coeff:  ").StoreASCII(material.mShiningStrengthCoeff)
		.StoreASCII("\nSelf Illum Coeff:        ").StoreASCII(material.mSelfIllumCoeff)
		.StoreASCII("\nOpacity Coeff:           ").StoreASCII(material.mOpacityCoeff)
		.StoreASCII("\nFilter Coeff:            ").StoreASCII(material.mFilterCoeff)
		.StoreASCII("\nBump Coeff:              ").StoreASCII(material.mBumpCoeff)
		.StoreASCII("\nReflexion Coeff:         ").StoreASCII(material.mReflexionCoeff)
		.StoreASCII("\nRefraction Coeff:        ").StoreASCII(material.mRefractionCoeff)
		.StoreASCII("\nDisplacement Coeff:      ").StoreASCII(material.mDisplacementCoeff)

		.StoreASCII("\nAmbient Color:           ").StoreASCII(material.mMtlAmbientColor.r).StoreASCII(" ").StoreASCII(material.mMtlAmbientColor.g).StoreASCII(" ").StoreASCII(material.mMtlAmbientColor.b)
		.StoreASCII("\nDiffuse Color:           ").StoreASCII(material.mMtlDiffuseColor.r).StoreASCII(" ").StoreASCII(material.mMtlDiffuseColor.g).StoreASCII(" ").StoreASCII(material.mMtlDiffuseColor.b)
		.StoreASCII("\nSpecular Color:          ").StoreASCII(material.mMtlSpecularColor.r).StoreASCII(" ").StoreASCII(material.mMtlSpecularColor.g).StoreASCII(" ").StoreASCII(material.mMtlSpecularColor.b)
		.StoreASCII("\nFilter Color:            ").StoreASCII(material.mMtlFilterColor.r).StoreASCII(" ").StoreASCII(material.mMtlFilterColor.g).StoreASCII(" ").StoreASCII(material.mMtlFilterColor.b)

		.StoreASCII("\nShading:                 ").StoreASCII(material.mShading)
		.StoreASCII("\nSoften:                  ").StoreASCII(material.mSoften)
		.StoreASCII("\nFaceMap:                 ").StoreASCII(material.mFaceMap)
		.StoreASCII("\nTwoSided:                ").StoreASCII(material.mTwoSided)
		.StoreASCII("\nWire:                    ").StoreASCII(material.mWire)
		.StoreASCII("\nWireUnits:               ").StoreASCII(material.mWireUnits)
		.StoreASCII("\nFalloffOut:              ").StoreASCII(material.mFalloffOut)
		.StoreASCII("\nTransparency type:       ").StoreASCII(material.mTransparencyType)

		.StoreASCII("\nShininess:               ").StoreASCII(material.mShininess)
		.StoreASCII("\nShiningStrength:         ").StoreASCII(material.mShiningStrength)
		.StoreASCII("\nSelfIllum:               ").StoreASCII(material.mSelfIllum)
		.StoreASCII("\nOpacity:                 ").StoreASCII(material.mOpacity)
		.StoreASCII("\nOpaFalloff:              ").StoreASCII(material.mOpaFalloff)
		.StoreASCII("\nWireSize:                ").StoreASCII(material.mWireSize)
		.StoreASCII("\nIOR:                     ").StoreASCII(material.mIOR)

		.StoreASCII("\nBounce:                  ").StoreASCII(material.mBounce)
		.StoreASCII("\nStatic Friction:         ").StoreASCII(material.mStaticFriction)
		.StoreASCII("\nSliding Friction:        ").StoreASCII(material.mSlidingFriction)

		.StoreASCII("\n");

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Mesh export method.
 *	This method is called once for each exported mesh.
 *	\param		mesh		a structure filled with current mesh information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportMesh(MeshInfo& mesh)
{
	// Export common information
	mGeomObjects.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("GeomObject:   ");
	ExportBasicInfo(&mesh, &mGeomObjects);

	// Export mesh parameters
	if(mesh.mIsCollapsed)		mGeomObjects.StoreASCII("(Collapsed object)\n");
	if(mesh.mIsSkeleton)		mGeomObjects.StoreASCII("(Skeleton object)\n");
	if(mesh.mIsInstance)		mGeomObjects.StoreASCII("(Instance object)\n");
	if(mesh.mIsTarget)			mGeomObjects.StoreASCII("(Target object)\n");
	if(!mesh.mIsConvertible)	mGeomObjects.StoreASCII("(Non convertible object)\n");
	if(mesh.mIsSkin)			mGeomObjects.StoreASCII("(Skin)\n");

	if(mesh.mIsSkeleton)
	{
		mGeomObjects.StoreASCII("CharID: ").StoreASCII(mesh.mCharID).StoreASCII("\n");
		mGeomObjects.StoreASCII("CSID:   ").StoreASCII(mesh.mCSID).StoreASCII("\n");
	}

	if(mesh.mIsInstance)
	{
		mGeomObjects.StoreASCII("Instance from: ").StoreASCII((const char*)mesh.mMasterMesh->mName).StoreASCII("\n");
	}
	else
	{
		// Export MAX native mesh
		mGeomObjects.StoreASCII("\nMAX native mesh data:\n");
		MAXNativeMesh* Mesh = &mesh.mOriginalMesh;

		mGeomObjects.StoreASCII(Mesh->mNbFaces).StoreASCII(" faces\n");
		mGeomObjects.StoreASCII(Mesh->mNbVerts).StoreASCII(" vertices\n");
		mGeomObjects.StoreASCII(Mesh->mNbTVerts).StoreASCII(" mapping coordinates\n");
		mGeomObjects.StoreASCII(Mesh->mNbCVerts).StoreASCII(" vertex colors\n");

		if(Mesh->mParity)	mGeomObjects.StoreASCII("Parity is true.\n");
		else				mGeomObjects.StoreASCII("Parity is false.\n");

		// Export vertices
		if(Mesh->mVerts)
		{
			mGeomObjects.StoreASCII("\nVertices:\n");
			MAXPoint* v = Mesh->mVerts;
			for(udword i=0;i<Mesh->mNbVerts;i++)
			{
				mGeomObjects.StoreASCII(v[i].x).StoreASCII(" ").StoreASCII(v[i].y).StoreASCII(" ").StoreASCII(v[i].z).StoreASCII("\n");
			}
		}

		// If the mesh is a skin, export skin-related data
		if(mesh.mIsSkin)
		{
			// Export links to the driving skeleton
			if(!Mesh->mBonesNb)
			{
				// A simple skin with one bone/vertex.
				MAXPoint* v = Mesh->mOffsetVectors;
				if(v)
				{
					mGeomObjects.StoreASCII("\nOffset vectors:\n");
					for(udword i=0;i<Mesh->mNbVerts;i++)
					{
						mGeomObjects.StoreASCII(v[i].x).StoreASCII(" ").StoreASCII(v[i].y).StoreASCII(" ").StoreASCII(v[i].z).StoreASCII("\n");
					}
				}

				if(Mesh->mBonesID)
				{
					mGeomObjects.StoreASCII("\nBones ID:\n");
					for(udword i=0;i<Mesh->mNbVerts;i++)
					{
						mGeomObjects.StoreASCII(Mesh->mBonesID[i]).StoreASCII(" ");
						if((i&63)==63)	mGeomObjects.StoreASCII("\n");
					}
					mGeomObjects.StoreASCII("\n");
				}
			}
			else
			{
				// A skin with multiple bones/vertex.
				mGeomObjects.StoreASCII("\nBones Nb:\n");
				udword Sum = 0;
				for(udword i=0;i<Mesh->mNbVerts;i++)
				{
					mGeomObjects.StoreASCII(Mesh->mBonesNb[i]).StoreASCII(" ");
					if((i&63)==63)	mGeomObjects.StoreASCII("\n");
					Sum+=Mesh->mBonesNb[i];
				}
				mGeomObjects.StoreASCII("\n");

				if(Mesh->mBonesID)
				{
					mGeomObjects.StoreASCII("\nBones ID:\n");
					for(udword i=0;i<Sum;i++)
					{
						mGeomObjects.StoreASCII(Mesh->mBonesID[i]).StoreASCII(" ");
						if((i&63)==63)	mGeomObjects.StoreASCII("\n");
					}
				}
				mGeomObjects.StoreASCII("\n");

				if(Mesh->mWeights)
				{
					mGeomObjects.StoreASCII("\nWeights:\n");
					for(udword i=0;i<Sum;i++)
					{
						mGeomObjects.StoreASCII(Mesh->mWeights[i]).StoreASCII(" ");
						if((i&63)==63)	mGeomObjects.StoreASCII("\n");
					}
				}
				mGeomObjects.StoreASCII("\n");

				if(Mesh->mOffsetVectors)
				{
					MAXPoint* v = Mesh->mOffsetVectors;
					mGeomObjects.StoreASCII("\nOffset vectors:\n");
					for(udword i=0;i<Sum;i++)
					{
						mGeomObjects.StoreASCII(v[i].x).StoreASCII(" ").StoreASCII(v[i].y).StoreASCII(" ").StoreASCII(v[i].z).StoreASCII("\n");
					}
				}
				mGeomObjects.StoreASCII("\n");
			}

			// Export compatible skeletal information
			Skeleton* Skel = Mesh->mSkeleton;
			mGeomObjects.StoreASCII("\nSkeletal information: ").StoreASCII(Skel->mNbBones).StoreASCII(" bones:\n");
			mGeomObjects.StoreASCII("(CSID, parent CSID)\n");
			udword* ID = Skel->mID;
			for(udword i=0;i<Skel->mNbBones;i++)
			{
				udword CSID = *ID++;
				udword pCSID = *ID++;
				mGeomObjects.StoreASCII(CSID).StoreASCII(" ").StoreASCII(pCSID).StoreASCII("\n");
			}
		}

		// Export UVW mappings
		if(Mesh->mTVerts)
		{
			mGeomObjects.StoreASCII("\nMapping coordinates:\n");
			MAXPoint* v = Mesh->mTVerts;
			for(udword i=0;i<Mesh->mNbTVerts;i++)
			{
				mGeomObjects.StoreASCII(v[i].x).StoreASCII(" ").StoreASCII(v[i].y).StoreASCII(" ").StoreASCII(v[i].z).StoreASCII("\n");
			}
		}

		// Export vertex-colors
		if(Mesh->mCVerts)
		{
			mGeomObjects.StoreASCII("\nVertex colors:\n");
			MAXPoint* v = Mesh->mCVerts;
			for(udword i=0;i<Mesh->mNbCVerts;i++)
			{
				mGeomObjects.StoreASCII(v[i].x).StoreASCII(" ").StoreASCII(v[i].y).StoreASCII(" ").StoreASCII(v[i].z).StoreASCII("\n");
			}
		}

		// Export faces
		if(Mesh->mFaces)
		{
			mGeomObjects.StoreASCII("\nFaces:\n(vref0 vref1 vref2  tref0 tref1 tref2  cref0 cref1 cref2  MatID Smg EdgeVisCode)\n");
			MAXFace* f = Mesh->mFaces;
			for(udword i=0;i<Mesh->mNbFaces;i++)
			{
				mGeomObjects
					.StoreASCII(f[i].VRef[0]).StoreASCII(" ").StoreASCII(f[i].VRef[1]).StoreASCII(" ").StoreASCII(f[i].VRef[2]).StoreASCII("  ")
					.StoreASCII(f[i].TRef[0]).StoreASCII(" ").StoreASCII(f[i].TRef[1]).StoreASCII(" ").StoreASCII(f[i].TRef[2]).StoreASCII("  ")
					.StoreASCII(f[i].CRef[0]).StoreASCII(" ").StoreASCII(f[i].CRef[1]).StoreASCII(" ").StoreASCII(f[i].CRef[2]).StoreASCII("  ")
					.StoreASCII(f[i].MatID).StoreASCII(" ").StoreASCII(f[i].Smg).StoreASCII(" ").StoreASCII(f[i].EdgeVis).StoreASCII("\n");
			}
		}

		// Export the convex hull
		if(Mesh->mFlags&MESH_CONVEXHULL)
		{
			mGeomObjects.StoreASCII("\nConvex hull:");
			mGeomObjects.StoreASCII("\n").StoreASCII(Mesh->mConvexHull->mNbVerts).StoreASCII(" vertices");
			mGeomObjects.StoreASCII("\n").StoreASCII(Mesh->mConvexHull->mNbFaces).StoreASCII(" faces\n");

			if(Mesh->mConvexHull->mVerts)
			{
				mGeomObjects.StoreASCII("\nConvex hull vertices:\n");
				MAXPoint* v = Mesh->mConvexHull->mVerts;
				for(udword i=0;i<Mesh->mConvexHull->mNbVerts;i++)
				{
					mGeomObjects.StoreASCII(v[i].x).StoreASCII(" ").StoreASCII(v[i].y).StoreASCII(" ").StoreASCII(v[i].z).StoreASCII("\n");
				}
			}

			if(Mesh->mConvexHull->mFaces)
			{
				mGeomObjects.StoreASCII("\nConvex hull faces:\n(vref0 vref1 vref2)\n");
				udword* f = Mesh->mConvexHull->mFaces;
				for(udword i=0;i<Mesh->mConvexHull->mNbFaces;i++)
				{
					mGeomObjects.StoreASCII(f[i*3+0]).StoreASCII(" ").StoreASCII(f[i*3+1]).StoreASCII(" ").StoreASCII(f[i*3+2]).StoreASCII("\n");
				}
			}
		}

		// Export the bounding sphere
		if(Mesh->mFlags&MESH_BOUNDINGSPHERE)
		{
			mGeomObjects
				.StoreASCII("\nBounding sphere:")
				.StoreASCII("\nCenter: ").StoreASCII(Mesh->mBSCenter.x).StoreASCII(" ").StoreASCII(Mesh->mBSCenter.y).StoreASCII(" ").StoreASCII(Mesh->mBSCenter.z)
				.StoreASCII("\nRadius: ").StoreASCII(Mesh->mBSRadius).StoreASCII("\n");
		}

		// Export the inertia tensor
		if(Mesh->mFlags&MESH_INERTIATENSOR)
		{
			ITensor* T = Mesh->mTensor;
			mGeomObjects
				.StoreASCII("\nVolume integrals:")

				.StoreASCII("\nCOM:  ").StoreASCII(T->COM.x).StoreASCII(" ").StoreASCII(T->COM.y).StoreASCII(" ").StoreASCII(T->COM.z)

				.StoreASCII("\nMass: ").StoreASCII(T->Mass).StoreASCII("\n")

				.StoreASCII("\nInertia tensor:\n")
				.StoreASCII(T->InertiaTensor[0][0]).StoreASCII(" ").StoreASCII(T->InertiaTensor[0][1]).StoreASCII(" ").StoreASCII(T->InertiaTensor[0][2]).StoreASCII("\n")
				.StoreASCII(T->InertiaTensor[1][0]).StoreASCII(" ").StoreASCII(T->InertiaTensor[1][1]).StoreASCII(" ").StoreASCII(T->InertiaTensor[1][2]).StoreASCII("\n")
				.StoreASCII(T->InertiaTensor[2][0]).StoreASCII(" ").StoreASCII(T->InertiaTensor[2][1]).StoreASCII(" ").StoreASCII(T->InertiaTensor[2][2]).StoreASCII("\n")

				.StoreASCII("\nCOM Inertia tensor:\n")
				.StoreASCII(T->COMInertiaTensor[0][0]).StoreASCII(" ").StoreASCII(T->COMInertiaTensor[0][1]).StoreASCII(" ").StoreASCII(T->COMInertiaTensor[0][2]).StoreASCII("\n")
				.StoreASCII(T->COMInertiaTensor[1][0]).StoreASCII(" ").StoreASCII(T->COMInertiaTensor[1][1]).StoreASCII(" ").StoreASCII(T->COMInertiaTensor[1][2]).StoreASCII("\n")
				.StoreASCII(T->COMInertiaTensor[2][0]).StoreASCII(" ").StoreASCII(T->COMInertiaTensor[2][1]).StoreASCII(" ").StoreASCII(T->COMInertiaTensor[2][2]).StoreASCII("\n");
		}

		// Export consolidated mesh
		if(Mesh->mFlags&MESH_CONSOLIDATION)
		{
			mGeomObjects.StoreASCII("\nCONSOLIDATION\n");

			// Topology
			{
				MESHBUILDERTOPOLOGY& Topo = mesh.mCleanMesh->Topology;

				mGeomObjects
				.StoreASCII("\nTOPOLOGY\n")
				.StoreASCII("\nNb faces:      ").StoreASCII(Topo.NbFaces)
				.StoreASCII("\nNb submeshes:  ").StoreASCII(Topo.NbSubmeshes)
				.StoreASCII("\n");

				// Submeshes
				for(udword i=0;i<Topo.NbSubmeshes;i++)
				{
					MESHBUILDERSUBMESH* CurSM = &Topo.SubmeshProperties[i];
					mGeomObjects
					.StoreASCII("\nSubmesh #").StoreASCII(i)
					.StoreASCII("\nMaterial ID:   ").StoreASCII(CurSM->MatID)
					.StoreASCII("\nSmGrps:        ").StoreASCII(CurSM->SmGrp)
					.StoreASCII("\nNbFaces:       ").StoreASCII(CurSM->NbFaces)
					.StoreASCII("\nNbVerts:       ").StoreASCII(CurSM->NbVerts)
					.StoreASCII("\nNbSubstrips:   ").StoreASCII(CurSM->NbSubstrips)
					.StoreASCII("\n");
				}

				// Connectivity
				uword* VRefs = Topo.VRefs;
				for(i=0;i<Topo.NbSubmeshes;i++)
				{
					uword NbFaces = Topo.FacesInSubmesh[i];

					mGeomObjects.StoreASCII("\nSubmesh #").StoreASCII(i).StoreASCII(" (").StoreASCII(NbFaces).StoreASCII(" faces)\n");

					// Save faces (Ref0, Ref1, Ref2)
					for(uword j=0;j<NbFaces;j++)
					{
						uword Ref0 = *VRefs++;
						uword Ref1 = *VRefs++;
						uword Ref2 = *VRefs++;
						mGeomObjects.StoreASCII(Ref0).StoreASCII(" ").StoreASCII(Ref1).StoreASCII(" ").StoreASCII(Ref2).StoreASCII("\n");
					}
				}

				// Face normals
				mGeomObjects.StoreASCII("\nFace normals:\n");
				float* Normals = Topo.Normals;
				if(Normals)
				{
					for(i=0;i<Topo.NbFaces;i++)
					{
						float nx = Normals[i*3+0];
						float ny = Normals[i*3+1];
						float nz = Normals[i*3+2];
						mGeomObjects.StoreASCII(nx).StoreASCII(" ").StoreASCII(ny).StoreASCII(" ").StoreASCII(nz).StoreASCII("\n");
					}
				}
				else mGeomObjects.StoreASCII("(not available)\n");
			}

			// Geometry
			{
				MESHBUILDERGEOMETRY& Geo = mesh.mCleanMesh->Geometry;

				mGeomObjects
				.StoreASCII("\nGEOMETRY\n")
				.StoreASCII("\nNb GeomPts:    ").StoreASCII(Geo.NbGeomPts)
				.StoreASCII("\nNb Verts:      ").StoreASCII(Geo.NbVerts)
				.StoreASCII("\nNb TVerts:     ").StoreASCII(Geo.NbTVerts)
				.StoreASCII("\n");

				// Indexed geometry
				udword MaxRef=0;
				if(Geo.VertsRefs)
				{
					mGeomObjects.StoreASCII("\nIndexed geometry:\n");
					for(udword i=0;i<Geo.NbVerts;i++)
					{
						udword Ref = Geo.VertsRefs[i];
						if(Ref>MaxRef)	MaxRef=Ref;
						mGeomObjects.StoreASCII(Ref).StoreASCII(" ");
						if((i&63)==63)	mGeomObjects.StoreASCII("\n");
					}
					mGeomObjects.StoreASCII("\n");
				}

				// Vertices
				udword NbVerts = Geo.VertsRefs ? (MaxRef+1) : Geo.NbVerts;
				mGeomObjects.StoreASCII("\nVertices: (").StoreASCII(NbVerts).StoreASCII(")\n");
				for(udword i=0;i<NbVerts;i++)
				{
					float x = Geo.Verts[i*3+0];
					float y = Geo.Verts[i*3+1];
					float z = Geo.Verts[i*3+2];
					mGeomObjects.StoreASCII(x).StoreASCII(" ").StoreASCII(y).StoreASCII(" ").StoreASCII(z).StoreASCII(" ").StoreASCII("\n");
				}
				mGeomObjects.StoreASCII("\n");

				// Indexed UVWs
				MaxRef = 0;
				if(Geo.TVertsRefs)
				{
					mGeomObjects.StoreASCII("\nIndexed UVWs:\n");
					for(udword i=0;i<Geo.NbVerts;i++)
					{
						udword Ref = Geo.TVertsRefs[i];
						if(Ref>MaxRef)	MaxRef=Ref;
						mGeomObjects.StoreASCII(Ref).StoreASCII(" ");
						if((i&63)==63)	mGeomObjects.StoreASCII("\n");
					}
					mGeomObjects.StoreASCII("\n");
				}

				// Vertices
				if(mSettings.ExpUVW && Geo.TVerts)
				{
					udword NbTVerts = Geo.TVertsRefs ? (MaxRef+1) : Geo.NbVerts;
					mGeomObjects.StoreASCII("\nUVW mappings: (").StoreASCII(NbTVerts).StoreASCII(")\n");
					float* p = Geo.TVerts;
					bool UseW = !mSettings.DiscardW;
					for(udword i=0;i<NbTVerts;i++)
					{
						float u = *p++;	mGeomObjects.StoreASCII(u).StoreASCII(" ");
						float v = *p++;	mGeomObjects.StoreASCII(v).StoreASCII(" ");
						if(UseW)
						{
							float w = *p++;	mGeomObjects.StoreASCII(w).StoreASCII(" ");
						}
						mGeomObjects.StoreASCII("\n");
					}
				}

				// Normals
				if(Geo.Normals)
				{
					udword NbNormals = Geo.NbVerts;
					mGeomObjects.StoreASCII("\nVertex normals: (").StoreASCII(NbNormals).StoreASCII(")\n");
					for(udword i=0;i<NbNormals;i++)
					{
						float x = Geo.Normals[i*3+0];
						float y = Geo.Normals[i*3+1];
						float z = Geo.Normals[i*3+2];
						mGeomObjects.StoreASCII(x).StoreASCII(" ").StoreASCII(y).StoreASCII(" ").StoreASCII(z).StoreASCII(" ").StoreASCII("\n");
					}
				}
				else mGeomObjects.StoreASCII("\nVertex normals: (not available)\n");

				// Vertex colors
				if(mSettings.ExpVTXColor && Geo.CVerts)
				{
					udword NbVtxColors = Geo.NbVerts;
					mGeomObjects.StoreASCII("\nVertex colors: (").StoreASCII(NbVtxColors).StoreASCII(")\n");
					for(udword i=0;i<NbVtxColors;i++)
					{
						float r = Geo.Normals[i*3+0];
						float g = Geo.Normals[i*3+1];
						float b = Geo.Normals[i*3+2];
						mGeomObjects.StoreASCII(r).StoreASCII(" ").StoreASCII(g).StoreASCII(" ").StoreASCII(b).StoreASCII(" ").StoreASCII("\n");
					}
				}

				// Normal info
				if(mSettings.ExportNormalInfo && Geo.NormalInfo)
				{
					udword NormalInfoSize = Geo.NormalInfoSize;
					mGeomObjects.StoreASCII("\nNormalInfo: (").StoreASCII(NormalInfoSize).StoreASCII(")\n");
					for(udword i=0;i<NormalInfoSize;i++)
					{
						udword d = Geo.NormalInfo[i];
						mGeomObjects.StoreASCII(d).StoreASCII(" ");
						if((i&63)==63)	mGeomObjects.StoreASCII("\n");
					}
				}
			}

			// Materials
			{
				MESHBUILDERMATERIALS& Mtl = mesh.mCleanMesh->Materials;

				mGeomObjects
				.StoreASCII("\nMATERIALS\n")
				.StoreASCII("\nNb materials:  ").StoreASCII(Mtl.NbMtls)
				.StoreASCII("\n");

				for(udword i=0;i<Mtl.NbMtls;i++)
				{
					MESHBUILDERMATINFO* CurMtl = &Mtl.MaterialInfo[i];

					mGeomObjects
					.StoreASCII("\nMaterial #").StoreASCII(i).StoreASCII(":")
					.StoreASCII("\nID:          ").StoreASCII(CurMtl->MatID)
					.StoreASCII("\nNbFaces:     ").StoreASCII(CurMtl->NbFaces)
					.StoreASCII("\nNbVerts:     ").StoreASCII(CurMtl->NbVerts)
					.StoreASCII("\nNbSubmeshes: ").StoreASCII(CurMtl->NbSubmeshes)
					.StoreASCII("\n");
				}
			}
		}
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Motion export method.
 *	This method is called once for each exported motion.
 *	\param		motion		a structure filled with current motion information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportMotion(MotionInfo& motion)
{
	// Export database management information
	mMotion.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("Motion for character: ");
	mMotion.StoreASCII(motion.mLinkID).StoreASCII("\n");

	mMotion.StoreASCII("Number of bones:  ").StoreASCII(motion.mNbBones).StoreASCII("\n");
	mMotion.StoreASCII("Number of vbones: ").StoreASCII(motion.mNbVirtualBones).StoreASCII("\n");

	udword* Data = (udword*)motion.mData;
	for(udword i=0;i<motion.mNbBones;i++)
	{
		udword CSID = *Data++;
		udword NbFrames = *Data++;
		mMotion.StoreASCII("CSID:    ").StoreASCII(CSID).StoreASCII("\n");
		mMotion.StoreASCII("#frames: ").StoreASCII(NbFrames).StoreASCII("\n");
		mMotion.StoreASCII("(tx, ty, tz)\n(qx qy qz qw)\n");
		float* p = (float*)Data;
		for(udword j=0;j<NbFrames;j++)
		{
			float tx = *p++;
			float ty = *p++;
			float tz = *p++;
			float qx = *p++;
			float qy = *p++;
			float qz = *p++;
			float qw = *p++;
			mMotion.StoreASCII(tx).StoreASCII(" ").StoreASCII(ty).StoreASCII(" ").StoreASCII(tz).StoreASCII("\n");
			mMotion.StoreASCII(qx).StoreASCII(" ").StoreASCII(qy).StoreASCII(" ").StoreASCII(qz).StoreASCII(" ").StoreASCII(qw).StoreASCII("\n");
		}
		Data = (udword*)p;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Shape export method.
 *	This method is called once for each exported shape.
 *	\param		shape		a structure filled with current shape information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportShape(ShapeInfo& shape)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Texture export method.
 *	This method is called once for each exported texture.
 *	\param		texture		a structure filled with current texture information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::ExportTexture(TextureInfo& texture)
{
	// Export database management information
	mTexmaps.StoreASCII("\n///////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n").StoreASCII("Texture: ");
	mTexmaps.StoreASCII((const char*)texture.mName);

	// We only export the texture's name in ascii, as well as the texture ID as a reminder.
	mTexmaps.StoreASCII("\n\nTextureID:    ").StoreASCII(texture.mObjectID);

	// The cropping values
	mTexmaps
		.StoreASCII("\n\nCropping values:\n")
		.StoreASCII("OffsetU: ").StoreASCII(texture.mCValues.OffsetU)
		.StoreASCII("\nOffsetV: ").StoreASCII(texture.mCValues.OffsetV)
		.StoreASCII("\nScaleU: ").StoreASCII(texture.mCValues.ScaleU)
		.StoreASCII("\nScaleV: ").StoreASCII(texture.mCValues.ScaleV)
		.StoreASCII("\n");

	// The texture matrix
	mTexmaps
		.StoreASCII("\nTexture matrix:\n")
		.StoreASCII(texture.mTMtx.m[0][0]).StoreASCII(" ").StoreASCII(texture.mTMtx.m[0][1]).StoreASCII(" ").StoreASCII(texture.mTMtx.m[0][2]).StoreASCII("\n")
		.StoreASCII(texture.mTMtx.m[1][0]).StoreASCII(" ").StoreASCII(texture.mTMtx.m[1][1]).StoreASCII(" ").StoreASCII(texture.mTMtx.m[1][2]).StoreASCII("\n")
		.StoreASCII(texture.mTMtx.m[2][0]).StoreASCII(" ").StoreASCII(texture.mTMtx.m[2][1]).StoreASCII(" ").StoreASCII(texture.mTMtx.m[2][2]).StoreASCII("\n")
		.StoreASCII(texture.mTMtx.m[3][0]).StoreASCII(" ").StoreASCII(texture.mTMtx.m[3][1]).StoreASCII(" ").StoreASCII(texture.mTMtx.m[3][2]).StoreASCII("\n\n");

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	End of export notification.
 *	This method is called once all nodes have been exported. This is a convenient place to free all used ram, etc.
 *	\param		stats		a structure filled with some export statistics.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ASCIIFormat::EndExport(StatsInfo& stats)
{
	// Complete general chunk
	if(!mIsMotionFile)
	{
		mGeneral
			.StoreASCII("Number of geom objects:      ")	.StoreASCII(stats.mNbGeomObjects)
			.StoreASCII("\nNumber of derived objects:   ")	.StoreASCII(stats.mNbDerivedObjects)
			.StoreASCII("\nNumber of cameras:           ")	.StoreASCII(stats.mNbCameras)
			.StoreASCII("\nNumber of lights:            ")	.StoreASCII(stats.mNbLights)
			.StoreASCII("\nNumber of shapes:            ")	.StoreASCII(stats.mNbShapes)
			.StoreASCII("\nNumber of helpers:           ")	.StoreASCII(stats.mNbHelpers)
			.StoreASCII("\nNumber of controllers:       ")	.StoreASCII(stats.mNbControllers)
			.StoreASCII("\nNumber of materials:         ")	.StoreASCII(stats.mNbMaterials)
			.StoreASCII("\nNumber of texmaps:           ")	.StoreASCII(stats.mNbTexmaps)
			.StoreASCII("\nNumber of unknown nodes:     ")	.StoreASCII(stats.mNbUnknowns)
			.StoreASCII("\nNumber of object-free nodes: ")	.StoreASCII(stats.mNbInvalidNodes)
			.StoreASCII("\n");
	}

	// Export to disk
	{
		// Open output file
		udword fp = CreateFile(mFilename, FILE_WRITE, FILE_CREATE_ALWAYS);
		if(fp!=-1)
		{
			// Write data to disk
			void* Data;
			Data = mGeneral			.Collapse();	if(Data)	WriteFile(fp, Data, mGeneral		.GetOffset());
			if(!mIsMotionFile)
			{
				Data = mGeomObjects		.Collapse();	if(Data)	WriteFile(fp, Data, mGeomObjects	.GetOffset());
				Data = mCameras			.Collapse();	if(Data)	WriteFile(fp, Data, mCameras		.GetOffset());
				Data = mLights			.Collapse();	if(Data)	WriteFile(fp, Data, mLights			.GetOffset());
				Data = mShapes			.Collapse();	if(Data)	WriteFile(fp, Data, mShapes			.GetOffset());
				Data = mHelpers			.Collapse();	if(Data)	WriteFile(fp, Data, mHelpers		.GetOffset());
				Data = mTexmaps			.Collapse();	if(Data)	WriteFile(fp, Data, mTexmaps		.GetOffset());
				Data = mMaterials		.Collapse();	if(Data)	WriteFile(fp, Data, mMaterials		.GetOffset());
				Data = mControllers		.Collapse();	if(Data)	WriteFile(fp, Data, mControllers	.GetOffset());
			}
			Data = mMotion			.Collapse();	if(Data)	WriteFile(fp, Data, mMotion			.GetOffset());

			// Close output file
			CloseFile(fp);
		}

		// Free everything for next call
		mGeneral		.Clean();
		mGeomObjects	.Clean();
		mCameras		.Clean();
		mLights			.Clean();
		mShapes			.Clean();
		mHelpers		.Clean();
		mTexmaps		.Clean();
		mMaterials		.Clean();
		mControllers	.Clean();
		mMotion			.Clean();
	}

	return true;
}

