///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	This file contains the ZCB Format exporter plug-in for Flexporter.
 *	\file		IceZCBExporter.cpp
 *	\author		Pierre Terdiman
 *	\date		April, 4, 2000
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Precompiled Header
#include "Stdafx.h"

static ZCBFormat		gZCBExporter;		// Gloabl exporter instance
static ExportSettings	gSettings;			// Global export settings

// FLEXPORTER Identification Callbacks

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Gives a brief exporter description. That string is displayed inside MAX, in the Options Panel.
 *	\relates	ZCBFormat
 *	\fn			ExporterDescription()
 *	\return		a description string
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT const char* ExporterDescription()
{ 
	return "ICE ZCB Exporter";
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Returns the format's extension. That string is displayed inside MAX, in the Options Panel.
 *	\relates	ZCBFormat
 *	\fn			FormatExtension()
 *	\return		an extension string
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT const char* FormatExtension()
{
	return "zcb";
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Returns the one and only exporter instance.
 *	\relates	ZCBFormat
 *	\fn			GetExporter()
 *	\return		pointer to the global exporter instance.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT ExportFormat* GetExporter()
{
	return &gZCBExporter;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Returns the default exportation settings for this format. This is the right place to initialize the default settings for your format.
 *	\relates	ZCBFormat
 *	\fn			GetDefaultSettings()
 *	\return		pointer to the global ExportSettings instance.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT ExportSettings* GetDefaultSettings()
{
	// General settings
	gSettings.Compressed					= false;
	gSettings.ExportWholeScene				= true;
	gSettings.ExportVisTrack				= false;
	gSettings.ExportHiddenNodes				= false;
	// Mesh settings
	gSettings.ExpUVW						= true;
	gSettings.DiscardW						= true;
	gSettings.ExpVTXColor					= true;
	gSettings.UseSmgrp						= true;
	gSettings.RemoveScaling					= true;
	gSettings.ConvertToD3D					= true;
	gSettings.AbsolutePRS					= false;
	gSettings.ConvexHull					= false;
	gSettings.BoundingSphere				= false;
	gSettings.InertiaTensor					= false;
	// Consolidation settings
	gSettings.ConsolidateMesh				= false;
	gSettings.ComputeFaceNormals			= false;
	gSettings.ComputeVertexNormals			= true;
	gSettings.ExportNormalInfo				= false;
	// Material settings
	gSettings.ForceAmbient					= false;
	gSettings.ForceDiffuse					= false;
	// Texture settings
	gSettings.OpacityInAlpha				= true;
	gSettings.TexnameOnly					= false;
	gSettings.KeepAmbientTexture			= false;
	gSettings.KeepDiffuseTexture			= true;
	gSettings.KeepSpecularTexture			= false;
	gSettings.KeepShininessTexture			= false;
	gSettings.KeepShiningStrengthTexture	= false;
	gSettings.KeepSelfIllumTexture			= false;
	gSettings.KeepOpacityTexture			= true;
	gSettings.KeepFilterTexture				= false;
	gSettings.KeepBumpTexture				= false;
	gSettings.KeepReflexionTexture			= false;
	gSettings.KeepRefractionTexture			= false;
	gSettings.KeepDisplacementTexture		= false;
	gSettings.TexMaxSize					= 256;
	// Camera settings
	gSettings.ExpCameras					= true;
	gSettings.ExpTargetNodes				= true;
	// Light settings
	gSettings.ExpLights						= true;
	// Animation settings
	gSettings.SingleFrame					= true;
	gSettings.Sampling						= true;
	gSettings.SamplingRate					= 1;

	return &gSettings;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Flexporter callback.
 *	Returns the FLEXPORTER SDK Version.
 *	\relates	ZCBFormat
 *	\fn			Version()
 *	\return		FLEXPORTER_VERSION
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
DLLEXPORT int Version()
{
	return FLEXPORTER_VERSION;
}









///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	A ZCB exporter plug-in for Flexporter..
 *	\class		ZCBFormat
 *	\author		Pierre Terdiman
 *	\version	1.0
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Constructor.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
ZCBFormat::ZCBFormat()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Destructor.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
ZCBFormat::~ZCBFormat()
{
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Plug-in initialization method.
 *	This method is called once before each export. When this method is called, the mSettings and mFilename members of the base format are valid.
 *	\param		true for pure motion files.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::Init(bool motion)
{
	// Header
	mGeneral.Store((char)'Z').Store((char)'C').Store((char)'B').Store((char)'!').Store((udword)motion ? ZCB_FILE_MOTION : ZCB_FILE_SCENE);

	// Chunk & version
	if(!motion)
	{
		mGeneral		.Store("MAIN").Store((udword)CHUNK_MAIN_VER);
		mGeomObjects	.Store("MESH").Store((udword)CHUNK_MESH_VER);
		mCameras		.Store("CAMS").Store((udword)CHUNK_CAMS_VER);
		mLights			.Store("LITE").Store((udword)CHUNK_LITE_VER);
		mShapes			.Store("SHAP").Store((udword)CHUNK_SHAP_VER);
		mHelpers		.Store("HELP").Store((udword)CHUNK_HELP_VER);
		mTexmaps		.Store("TEXM").Store((udword)CHUNK_TEXM_VER);
		mMaterials		.Store("MATL").Store((udword)CHUNK_MATL_VER);
		mControllers	.Store("CTRL").Store((udword)CHUNK_CTRL_VER);
		mMotion			.Store("MOVE").Store((udword)CHUNK_MOVE_VER);
	}
	else
	{
		mMotion			.Store("MOVE").Store((udword)CHUNK_MOVE_VER);
	}

	mIsMotionFile	= motion;

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Scene export method.
 *	This method is called once to export the scene global information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::SetSceneInfo(MainInfo& maininfo)
{
	// Save time info for later
	CopyMemory(&mTimeInfo, &maininfo.Time, sizeof(TimeInfo));

	// Export time info
	mGeneral
		.Store((long)(mTimeInfo.StartTime / mTimeInfo.DeltaTime))
		.Store((long)(mTimeInfo.EndTime / mTimeInfo.DeltaTime))
		.Store(mTimeInfo.FrameRate).Store((long)mTimeInfo.DeltaTime);

	// Export background color
	mGeneral.Store(maininfo.BackColor.r).Store(maininfo.BackColor.g).Store(maininfo.BackColor.b);

	// Export ambient color
	mGeneral.Store(maininfo.AmbColor.r).Store(maininfo.AmbColor.g).Store(maininfo.AmbColor.b);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Common method used to export basic information for a mesh, a camera, a light, etc.
 *	\param		obj			current object
 *	\param		array		destination export array
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportBasicInfo(ObjectInfo* obj, CustomArray* array)
{
	if(!array || !obj) return false;

	// Export name
	array	->Store((const char*)obj->mName).Store((ubyte)0);

	// Export database management information
	array	->Store(obj->mObjectID)
			.Store(obj->mParentID)
			.Store(obj->mLinkID)
			.Store(obj->mIsGroupMember);

	// Export common properties
	array	->Store(obj->mPrs.Position.x)
			.Store(obj->mPrs.Position.y)
			.Store(obj->mPrs.Position.z)

			.Store(obj->mPrs.Rotation.x)
			.Store(obj->mPrs.Rotation.y)
			.Store(obj->mPrs.Rotation.z)
			.Store(obj->mPrs.Rotation.w)

			.Store(obj->mPrs.Scale.x)
			.Store(obj->mPrs.Scale.y)
			.Store(obj->mPrs.Scale.z)

			.Store(obj->mWireColor)

			.Store(obj->mLocalPRS)
			.Store(obj->mD3DCompliant);

	// Export user-defined properties
	if(obj->mUserProps)	array->Store((const char*)obj->mUserProps);
	else				array->Store((ubyte)0);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Camera export method.
 *	This method is called once for each exported camera.
 *	\param		camera		a structure filled with current camera information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportCamera(CameraInfo& camera)
{
	// Export common information
	ExportBasicInfo(&camera, &mCameras);

	// Save parameters
	mCameras
		.Store(camera.mOrthoCam)
		.Store(camera.mFOV)
		.Store(camera.mNearClip)
		.Store(camera.mFarClip)
		.Store(camera.mTDist)
		.Store(camera.mHLineDisplay)
		.Store(camera.mEnvNearRange)
		.Store(camera.mEnvFarRange)
		.Store(camera.mEnvDisplay)
		.Store(camera.mManualClip);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Controller export method.
 *	This method is called once for each exported controller.
 *	\param		controller		a structure filled with current controller information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportController(ControllerInfo& controller)
{
	mControllers
		.Store((const char*)controller.mField).Store((ubyte)0)		// Controlled field
		.Store(controller.mObjectID)								// Controller ID
		.Store(controller.mOwnerID);								// Owner ID

	// Controller data
	ControllerData* cdata = controller.mData;

	udword NbValues;

			if(cdata->mType==CTRL_FLOAT)		{mControllers.Store((udword)ZCB_CTRL_FLOAT);	NbValues = 1; }
	else	if(cdata->mType==CTRL_VECTOR)		{mControllers.Store((udword)ZCB_CTRL_VECTOR);	NbValues = 3; }
	else	if(cdata->mType==CTRL_QUAT)			{mControllers.Store((udword)ZCB_CTRL_QUAT);		NbValues = 4; }
	else	if(cdata->mType==CTRL_PR)			{mControllers.Store((udword)ZCB_CTRL_PR);		NbValues = 7; }
	else	if(cdata->mType==CTRL_PRS)			{mControllers.Store((udword)ZCB_CTRL_PRS);		NbValues = 10; }
	else										{mControllers.Store((udword)ZCB_CTRL_NONE);		NbValues = 0; }

			if(cdata->mMode==CTRL_SAMPLES)		mControllers.Store((udword)ZCB_CTRL_SAMPLES);
	else	if(cdata->mMode==CTRL_KEYFRAMES)	mControllers.Store((udword)ZCB_CTRL_KEYFRAMES);
	else										mControllers.Store((udword)ZCB_CTRL_NONE);

	// Dump the samples
	if(cdata->mMode==CTRL_SAMPLES)
	{
		SampleData* sdata = (SampleData*)cdata;
		mControllers.Store(sdata->mNbSamples).Store(sdata->mSamplingRate);

		float* Values = (float*)sdata->mSamples;
		for(udword i=0;i<NbValues*sdata->mNbSamples;i++)
		{
			mControllers.Store(*Values++);
		}
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Helper export method.
 *	This method is called once for each exported helper.
 *	\param		helper		a structure filled with current helper information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportHelper(HelperInfo& helper)
{
	// Export common information
	ExportBasicInfo(&helper, &mHelpers);

	// Save parameters
	mHelpers.Store(helper.mIsGroupHead);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Light export method.
 *	This method is called once for each exported light.
 *	\param		light		a structure filled with current light information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportLight(LightInfo& light)
{
	// Export common information
	ExportBasicInfo(&light, &mLights);

	// Save parameters
	mLights
		.Store(light.mLightType)
		.Store(light.mIsSpot)
		.Store(light.mIsDir)
		.Store(light.mColor.x)
		.Store(light.mColor.y)
		.Store(light.mColor.z)
		.Store(light.mIntensity)
		.Store(light.mContrast)
		.Store(light.mDiffuseSoft)
		.Store(light.mLightUsed)
		.Store(light.mAffectDiffuse)
		.Store(light.mAffectSpecular)
		.Store(light.mUseAttenNear)
		.Store(light.mAttenNearDisplay)
		.Store(light.mUseAtten)
		.Store(light.mShowAtten)
		.Store(light.mNearAttenStart)
		.Store(light.mNearAttenEnd)
		.Store(light.mAttenStart)
		.Store(light.mAttenEnd)
		.Store(light.mDecayType)
		.Store(light.mHotSpot)
		.Store(light.mFallsize)
		.Store(light.mAspect)
		.Store(light.mSpotShape)
		.Store(light.mOvershoot)
		.Store(light.mConeDisplay)
		.Store(light.mTDist)
		.Store(light.mShadowType)
		.Store(light.mAbsMapBias)
		.Store(light.mRayBias)
		.Store(light.mMapBias)
		.Store(light.mMapRange)
		.Store(light.mMapSize);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Material export method.
 *	This method is called once for each exported material.
 *	\param		material		a structure filled with current material information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportMaterial(MaterialInfo& material)
{
	// Export name
	mMaterials.Store((const char*)material.mName).Store((ubyte)0);

	// Export material ID
	mMaterials.Store(material.mObjectID);

	// Export material parameters
	mMaterials
		.Store(material.mAmbientMapID)
		.Store(material.mDiffuseMapID)
		.Store(material.mSpecularMapID)
		.Store(material.mShininessMapID)
		.Store(material.mShiningStrengthMapID)
		.Store(material.mSelfIllumMapID)
		.Store(material.mOpacityMapID)
		.Store(material.mFilterMapID)
		.Store(material.mBumpMapID)
		.Store(material.mReflexionMapID)
		.Store(material.mRefractionMapID)
		.Store(material.mDisplacementMapID)

		.Store(material.mAmbientCoeff)
		.Store(material.mDiffuseCoeff)
		.Store(material.mSpecularCoeff)
		.Store(material.mShininessCoeff)
		.Store(material.mShiningStrengthCoeff)
		.Store(material.mSelfIllumCoeff)
		.Store(material.mOpacityCoeff)
		.Store(material.mFilterCoeff)
		.Store(material.mBumpCoeff)
		.Store(material.mReflexionCoeff)
		.Store(material.mRefractionCoeff)
		.Store(material.mDisplacementCoeff)

		.Store(material.mMtlAmbientColor.r)		.Store(material.mMtlAmbientColor.g)		.Store(material.mMtlAmbientColor.b)
		.Store(material.mMtlDiffuseColor.r)		.Store(material.mMtlDiffuseColor.g)		.Store(material.mMtlDiffuseColor.b)
		.Store(material.mMtlSpecularColor.r)	.Store(material.mMtlSpecularColor.g)	.Store(material.mMtlSpecularColor.b)
		.Store(material.mMtlFilterColor.r)		.Store(material.mMtlFilterColor.g)		.Store(material.mMtlFilterColor.b)

		.Store(material.mShading)
		.Store(material.mSoften)
		.Store(material.mFaceMap)
		.Store(material.mTwoSided)
		.Store(material.mWire)
		.Store(material.mWireUnits)
		.Store(material.mFalloffOut)
		.Store(material.mTransparencyType)

		.Store(material.mShininess)
		.Store(material.mShiningStrength)
		.Store(material.mSelfIllum)
		.Store(material.mOpacity)
		.Store(material.mOpaFalloff)
		.Store(material.mWireSize)
		.Store(material.mIOR)

		.Store(material.mBounce)
		.Store(material.mStaticFriction)
		.Store(material.mSlidingFriction);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Mesh export method.
 *	This method is called once for each exported mesh.
 *	\param		mesh		a structure filled with current mesh information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportMesh(MeshInfo& mesh)
{
	// Export common information
	ExportBasicInfo(&mesh, &mGeomObjects);

	// Export mesh parameters
	mGeomObjects
		.Store(mesh.mIsCollapsed)
		.Store(mesh.mIsSkeleton)
		.Store(mesh.mIsInstance)
		.Store(mesh.mIsTarget)
		.Store(mesh.mIsConvertible)
		.Store(mesh.mIsSkin);

	if(mesh.mIsSkeleton)	mGeomObjects.Store(mesh.mCharID).Store(mesh.mCSID);

	if(!mesh.mIsInstance)	// If that's an instance, use mLinkID
	{
		// Export MAX native mesh
		MAXNativeMesh* Mesh = &mesh.mOriginalMesh;

		mGeomObjects.Store(Mesh->mNbFaces).Store(Mesh->mNbVerts).Store(Mesh->mNbTVerts).Store(Mesh->mNbCVerts).Store(Mesh->mFlags).Store(Mesh->mParity);

		// Save vertices (null for skins)
		if(Mesh->mVerts)
		{
			MAXPoint* v = Mesh->mVerts;
			for(udword i=0;i<Mesh->mNbVerts;i++)		mGeomObjects.Store(v[i].x).Store(v[i].y).Store(v[i].z);
		}

		// Save skin-related data
		if(mesh.mIsSkin)
		{
			// Export links to the driving skeleton
			if(!Mesh->mBonesNb)
			{
				// A simple skin with one bone/vertex.

				// Save offset vectors
				MAXPoint* v = Mesh->mOffsetVectors;
				for(udword i=0;i<Mesh->mNbVerts;i++)	mGeomObjects.Store(v[i].x).Store(v[i].y).Store(v[i].z);

				// Save bones ID
				for(i=0;i<Mesh->mNbVerts;i++)			mGeomObjects.Store(Mesh->mBonesID[i]);
			}
			else
			{
				// A skin with multiple bones/vertex.

				// Save bones Nb
				udword Sum = 0;
				for(udword i=0;i<Mesh->mNbVerts;i++)
				{
					mGeomObjects.Store(Mesh->mBonesNb[i]);
					Sum+=Mesh->mBonesNb[i];
				}

				// Save bones ID
				for(i=0;i<Sum;i++)						mGeomObjects.Store(Mesh->mBonesID[i]);

				// Save weights
				for(i=0;i<Sum;i++)						mGeomObjects.Store(Mesh->mWeights[i]);

				// Save offset vectors
				MAXPoint* v = Mesh->mOffsetVectors;
				for(i=0;i<Sum;i++)						mGeomObjects.Store(v[i].x).Store(v[i].y).Store(v[i].z);
			}

			// Export compatible skeletal information
			Skeleton* Skel = Mesh->mSkeleton;
			mGeomObjects.Store(Skel->mNbBones);
			udword* ID = Skel->mID;
			for(udword i=0;i<Skel->mNbBones;i++)
			{
				udword CSID = *ID++;
				udword pCSID = *ID++;
				mGeomObjects.Store(CSID).Store(pCSID);
			}
		}

		// Save mapping coordinates
		if(Mesh->mTVerts)
		{
			MAXPoint* v = Mesh->mTVerts;
			for(udword i=0;i<Mesh->mNbTVerts;i++)
			{
				mGeomObjects.Store(v[i].x).Store(v[i].y);							// Save UVs
				if(!(Mesh->mFlags&MESH_WDISCARDED))	mGeomObjects.Store(v[i].z);		// Discard W if needed
			}
		}

		// Save vertex-colors
		if(Mesh->mCVerts)
		{
			MAXPoint* v = Mesh->mCVerts;
			for(udword i=0;i<Mesh->mNbCVerts;i++)			mGeomObjects.Store(v[i].x).Store(v[i].y).Store(v[i].z);
		}

		// Save faces
		MAXFace* f = Mesh->mFaces;
		for(udword i=0;i<Mesh->mNbFaces;i++)
		{
			if(Mesh->mFlags&MESH_VFACE)	mGeomObjects.Store(f[i].VRef[0])	.Store(f[i].VRef[1])	.Store(f[i].VRef[2]);
			if(Mesh->mFlags&MESH_TFACE)	mGeomObjects.Store(f[i].TRef[0])	.Store(f[i].TRef[1])	.Store(f[i].TRef[2]);
			if(Mesh->mFlags&MESH_CFACE)	mGeomObjects.Store(f[i].CRef[0])	.Store(f[i].CRef[1])	.Store(f[i].CRef[2]);
										mGeomObjects.Store(f[i].MatID)		.Store(f[i].Smg)		.Store(f[i].EdgeVis);
		}

		// Export the possible convex hull
		if(Mesh->mFlags&MESH_CONVEXHULL)
		{
			mGeomObjects.Store(Mesh->mConvexHull->mNbVerts).Store(Mesh->mConvexHull->mNbFaces);
			if(Mesh->mConvexHull->mVerts)
			{
				MAXPoint* v = Mesh->mConvexHull->mVerts;
				for(udword i=0;i<Mesh->mConvexHull->mNbVerts;i++)	mGeomObjects.Store(v[i].x).Store(v[i].y).Store(v[i].z);
			}

			if(Mesh->mConvexHull->mFaces)
			{
				udword* f = Mesh->mConvexHull->mFaces;
				for(udword i=0;i<Mesh->mConvexHull->mNbFaces;i++)	mGeomObjects.Store(f[i*3+0]).Store(f[i*3+1]).Store(f[i*3+2]);
			}
		}

		// Export the possible bounding sphere
		if(Mesh->mFlags&MESH_BOUNDINGSPHERE)
		{
			mGeomObjects.Store(Mesh->mBSCenter.x).Store(Mesh->mBSCenter.y).Store(Mesh->mBSCenter.z).Store(Mesh->mBSRadius);
		}

		// Export the inertia tensor
		if(Mesh->mFlags&MESH_INERTIATENSOR)
		{
			ITensor* T = Mesh->mTensor;
			mGeomObjects
				.Store(T->COM.x).Store(T->COM.y).Store(T->COM.z)
				.Store(T->Mass)
				.Store(T->InertiaTensor[0][0]).Store(T->InertiaTensor[0][1]).Store(T->InertiaTensor[0][2])
				.Store(T->InertiaTensor[1][0]).Store(T->InertiaTensor[1][1]).Store(T->InertiaTensor[1][2])
				.Store(T->InertiaTensor[2][0]).Store(T->InertiaTensor[2][1]).Store(T->InertiaTensor[2][2])
				.Store(T->COMInertiaTensor[0][0]).Store(T->COMInertiaTensor[0][1]).Store(T->COMInertiaTensor[0][2])
				.Store(T->COMInertiaTensor[1][0]).Store(T->COMInertiaTensor[1][1]).Store(T->COMInertiaTensor[1][2])
				.Store(T->COMInertiaTensor[2][0]).Store(T->COMInertiaTensor[2][1]).Store(T->COMInertiaTensor[2][2]);
		}

		// Export consolidated mesh
		if(Mesh->mFlags&MESH_CONSOLIDATION)
		{
			// Topology
			{
				MESHBUILDERTOPOLOGY& Topo = mesh.mCleanMesh->Topology;

				mGeomObjects.Store(Topo.NbFaces).Store(Topo.NbSubmeshes);

				// Submeshes
				for(udword i=0;i<Topo.NbSubmeshes;i++)
				{
					MESHBUILDERSUBMESH* CurSM = &Topo.SubmeshProperties[i];
					mGeomObjects.Store(CurSM->MatID).Store(CurSM->SmGrp).Store(CurSM->NbFaces).Store(CurSM->NbVerts).Store(CurSM->NbSubstrips);
				}

				// Connectivity
				uword* VRefs = Topo.VRefs;
				for(i=0;i<Topo.NbSubmeshes;i++)
				{
					uword NbFaces = Topo.FacesInSubmesh[i];

					mGeomObjects.Store(NbFaces);

					// Save faces (Ref0, Ref1, Ref2)
					for(uword j=0;j<NbFaces;j++)
					{
						uword Ref0 = *VRefs++;
						uword Ref1 = *VRefs++;
						uword Ref2 = *VRefs++;
						mGeomObjects.Store(Ref0).Store(Ref1).Store(Ref2);
					}
				}

				// Face normals
				if(Mesh->mFlags&MESH_FACENORMALS)
				{
					float* Normals = Topo.Normals;
					for(i=0;i<Topo.NbFaces;i++)
					{
						float nx = Normals ? Normals[i*3+0] : 0.0f;
						float ny = Normals ? Normals[i*3+1] : 0.0f;
						float nz = Normals ? Normals[i*3+2] : 0.0f;
						mGeomObjects.Store(nx).Store(ny).Store(nz);
					}
				}
			}

			// Geometry
			{
				MESHBUILDERGEOMETRY& Geo = mesh.mCleanMesh->Geometry;

				mGeomObjects.Store(Geo.NbGeomPts).Store(Geo.NbVerts).Store(Geo.NbTVerts);

				// Indexed geometry
				udword MaxRef=0;
				if(Geo.VertsRefs)
				{
					for(udword i=0;i<Geo.NbVerts;i++)
					{
						udword Ref = Geo.VertsRefs[i];
						if(Ref>MaxRef)	MaxRef=Ref;
						mGeomObjects.Store(Ref);
					}
				}

				// Vertices
				udword NbVerts = Geo.VertsRefs ? (MaxRef+1) : Geo.NbVerts;
				mGeomObjects.Store(NbVerts);
				for(udword i=0;i<NbVerts;i++)
				{
					float x = Geo.Verts[i*3+0];
					float y = Geo.Verts[i*3+1];
					float z = Geo.Verts[i*3+2];
					mGeomObjects.Store(x).Store(y).Store(z);
				}

				// Indexed UVWs
				MaxRef = 0;
				if((Mesh->mFlags&MESH_UVW) && Geo.TVertsRefs)
				{
					for(udword i=0;i<Geo.NbVerts;i++)
					{
						udword Ref = Geo.TVertsRefs[i];
						if(Ref>MaxRef)	MaxRef=Ref;
						mGeomObjects.Store(Ref);
					}
				}

				// Vertices
				if((Mesh->mFlags&MESH_UVW) && Geo.TVerts)
				{
					udword NbTVerts = Geo.TVertsRefs ? (MaxRef+1) : Geo.NbVerts;
					mGeomObjects.Store(NbTVerts);
					float* p = Geo.TVerts;
					for(udword i=0;i<NbTVerts;i++)
					{
						float u = *p++;	mGeomObjects.Store(u);
						float v = *p++;	mGeomObjects.Store(v);
						if(!(Mesh->mFlags&MESH_WDISCARDED))	
						{
							float w = *p++;	mGeomObjects.Store(w);
						}
					}
				}

				// Normals
				if(Mesh->mFlags&MESH_VERTEXNORMALS)
				{
					udword NbNormals = Geo.NbVerts;
					mGeomObjects.Store(NbNormals);
					for(udword i=0;i<NbNormals;i++)
					{
						float x = Geo.Normals[i*3+0];
						float y = Geo.Normals[i*3+1];
						float z = Geo.Normals[i*3+2];
						mGeomObjects.Store(x).Store(y).Store(z);
					}
				}

				// Vertex colors
				if((Mesh->mFlags&MESH_VERTEXCOLORS) && Geo.CVerts)
				{
					udword NbVtxColors = Geo.NbVerts;
					mGeomObjects.Store(NbVtxColors);
					for(udword i=0;i<NbVtxColors;i++)
					{
						float r = Geo.CVerts[i*3+0];
						float g = Geo.CVerts[i*3+1];
						float b = Geo.CVerts[i*3+2];
						mGeomObjects.Store(r).Store(g).Store(b);
					}
				}

				// Normal info
				if((Mesh->mFlags&MESH_NORMALINFO) && Geo.NormalInfo)
				{
					udword NormalInfoSize = Geo.NormalInfoSize;
					mGeomObjects.Store(NormalInfoSize);
					for(udword i=0;i<NormalInfoSize;i++)
					{
						udword d = Geo.NormalInfo[i];
						mGeomObjects.Store(d);
					}
				}
			}

			// Materials
			{
				MESHBUILDERMATERIALS& Mtl = mesh.mCleanMesh->Materials;

				mGeomObjects.Store(Mtl.NbMtls);

				for(udword i=0;i<Mtl.NbMtls;i++)
				{
					MESHBUILDERMATINFO* CurMtl = &Mtl.MaterialInfo[i];

					mGeomObjects.Store(CurMtl->MatID).Store(CurMtl->NbFaces).Store(CurMtl->NbVerts).Store(CurMtl->NbSubmeshes);
				}
			}
		}
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Motion export method.
 *	This method is called once for each exported motion.
 *	\param		motion		a structure filled with current motion information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportMotion(MotionInfo& motion)
{
	mMotion.Store(motion.mLinkID).Store(motion.mNbBones).Store(motion.mNbVirtualBones);

	udword* Data = (udword*)motion.mData;
	for(udword i=0;i<motion.mNbBones;i++)
	{
		udword CSID = *Data++;
		udword NbFrames = *Data++;
		mMotion.Store(CSID).Store(NbFrames);
		float* p = (float*)Data;
		for(udword j=0;j<NbFrames;j++)
		{
			float tx = *p++;
			float ty = *p++;
			float tz = *p++;
			float qx = *p++;
			float qy = *p++;
			float qz = *p++;
			float qw = *p++;
			mMotion
				.Store(tx).Store(ty).Store(tz)
				.Store(qx).Store(qy).Store(qz).Store(qw);
		}
		Data = (udword*)p;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Shape export method.
 *	This method is called once for each exported shape.
 *	\param		shape		a structure filled with current shape information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportShape(ShapeInfo& shape)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Texture export method.
 *	This method is called once for each exported texture.
 *	\param		texture		a structure filled with current texture information.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::ExportTexture(TextureInfo& texture)
{
	// Export name
	mTexmaps.Store((const char*)texture.mName).Store((ubyte)0);

	// Export texture ID
	mTexmaps.Store(texture.mObjectID);

	// Export texture data
	mTexmaps.Store(texture.mWidth).Store(texture.mHeight).Store(texture.mHasAlpha);
	ubyte* Pixels = texture.mPixels;
	for(udword i=0;i<texture.mWidth*texture.mHeight;i++)
	{
		ubyte R = *Pixels++;
		ubyte G = *Pixels++;
		ubyte B = *Pixels++;
		ubyte A = *Pixels++;
		mTexmaps.Store(R).Store(G).Store(B);
		if(texture.mHasAlpha)	mTexmaps.Store(A);
	}

	// Export cropping values
	mTexmaps
		.Store(texture.mCValues.OffsetU).Store(texture.mCValues.OffsetV).Store(texture.mCValues.ScaleU).Store(texture.mCValues.ScaleV);

	// Export texture matrix
	mTexmaps
		.Store(texture.mTMtx.m[0][0]).Store(texture.mTMtx.m[0][1]).Store(texture.mTMtx.m[0][2])
		.Store(texture.mTMtx.m[1][0]).Store(texture.mTMtx.m[1][1]).Store(texture.mTMtx.m[1][2])
		.Store(texture.mTMtx.m[2][0]).Store(texture.mTMtx.m[2][1]).Store(texture.mTMtx.m[2][2])
		.Store(texture.mTMtx.m[3][0]).Store(texture.mTMtx.m[3][1]).Store(texture.mTMtx.m[3][2]);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	End of export notification.
 *	This method is called once all nodes have been exported. This is a convenient place to free all used ram, etc.
 *	\param		stats		a structure filled with some export statistics.
 *	\return		true if success.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool ZCBFormat::EndExport(StatsInfo& stats)
{
	// Complete general chunk
	if(!mIsMotionFile)
		mGeneral
			.Store(stats.mNbGeomObjects)
			.Store(stats.mNbDerivedObjects)
			.Store(stats.mNbCameras)
			.Store(stats.mNbLights)
			.Store(stats.mNbShapes)
			.Store(stats.mNbHelpers)
			.Store(stats.mNbControllers)
			.Store(stats.mNbMaterials)
			.Store(stats.mNbTexmaps)
			.Store(stats.mNbUnknowns)
			.Store(stats.mNbInvalidNodes);

	// Export to disk
	{
		// Open output file
		udword fp = CreateFile(mFilename, FILE_WRITE, FILE_CREATE_ALWAYS);
		if(fp!=-1)
		{
			// Write data to disk
			void* Data;
				// The "General" array must be exported first, since it may include the format's signature.
											{ Data = mGeneral			.Collapse();	if(Data)	WriteFile(fp, Data, mGeneral		.GetOffset());}
			if(!mIsMotionFile)
			{
				if(stats.mNbGeomObjects)	{ Data = mGeomObjects		.Collapse();	if(Data)	WriteFile(fp, Data, mGeomObjects	.GetOffset());}
				if(stats.mNbCameras)		{ Data = mCameras			.Collapse();	if(Data)	WriteFile(fp, Data, mCameras		.GetOffset());}
				if(stats.mNbLights)			{ Data = mLights			.Collapse();	if(Data)	WriteFile(fp, Data, mLights			.GetOffset());}
				if(stats.mNbShapes)			{ Data = mShapes			.Collapse();	if(Data)	WriteFile(fp, Data, mShapes			.GetOffset());}
				if(stats.mNbHelpers)		{ Data = mHelpers			.Collapse();	if(Data)	WriteFile(fp, Data, mHelpers		.GetOffset());}
				if(stats.mNbTexmaps)		{ Data = mTexmaps			.Collapse();	if(Data)	WriteFile(fp, Data, mTexmaps		.GetOffset());}
				if(stats.mNbMaterials)		{ Data = mMaterials			.Collapse();	if(Data)	WriteFile(fp, Data, mMaterials		.GetOffset());}
				if(stats.mNbControllers)	{ Data = mControllers		.Collapse();	if(Data)	WriteFile(fp, Data, mControllers	.GetOffset());}
				if(stats.mNbDerivedObjects)	{ Data = mMotion			.Collapse();	if(Data)	WriteFile(fp, Data, mMotion			.GetOffset());}
			}
			else							{ Data = mMotion			.Collapse();	if(Data)	WriteFile(fp, Data, mMotion			.GetOffset());}

			// Close output file
			CloseFile(fp);
		}

		// Free everything for next call
		mGeneral		.Clean();
		mGeomObjects	.Clean();
		mCameras		.Clean();
		mLights			.Clean();
		mShapes			.Clean();
		mHelpers		.Clean();
		mTexmaps		.Clean();
		mMaterials		.Clean();
		mControllers	.Clean();
		mMotion			.Clean();
	}

	return true;
}

